
inline void project_SJS(DataSet* data,float* v, unsigned int *P, int z_dimensions,
        vector<float> rand_vector, vector<int> dimension_permutation, float elem_min,
        float elem_max,  float precision) {

    /*printf("MIN: %f, MAX: %f\n",elem_min,elem_max);*/
    for(int i_dim =0; i_dim < data->dimensionality; i_dim++)
    {
        int proj_dim = (i_dim % z_dimensions);
        if (i_dim < z_dimensions) P[proj_dim] = 0; //TODO: optimize
        float shift_val = 0;
        shift_val = rand_vector[i_dim];
        float coord = 0;
        /*coord = (( (coord - min_value)/(max_value-min_value) +shift_val)*precision);*/
        /*printf("MIN: %f, MAX: %f\n",elem_min,elem_max);*/
        coord = *(v + dimension_permutation[i_dim]);
        coord = (( (coord - elem_min)/(elem_max-elem_min) +shift_val)*precision);
#ifdef SANITYCHECK
        if(coord < 0 || P[proj_dim]+coord > UINT_MAX) {
            printf("MIN: %f, MAX: %f\n",elem_min,elem_max);
            printf("COORD:%f > %u (elem_min:%f, elem_max:%f)\n",P[proj_dim]+coord,UINT_MAX,elem_min,elem_max);
            terminal_error("Scaling to int failed\n");
        }
#endif
        P[proj_dim] += (unsigned int) coord;
    }
}

inline void project_mean_adjust(DataSet* data,float* v, unsigned int *P, int z_dimensions,
        vector<float> rand_vector, vector<int> dimension_permutation, float elem_min,
        float elem_max,  float precision) {

    for(int i_dim =0; i_dim < data->dimensionality; i_dim++)
    {
        float coord = 0;
        int proj_dim=i_dim;

        //max 32 bit uint: 4294967295
        //max 32 bit int: 2147483647

        coord = v[dimension_permutation[i_dim]];
        coord = coord + 4294967295/2.0; // Align mean to halfway of UINT range.
        /*coord = coord + UINT_MAX/2.0; // Align mean to halfway of UINT range.*/

        P[proj_dim] = (unsigned int) coord;

#ifdef SANITYCHECK
        if(coord < 0 || coord > UINT_MAX) {
            printf("MIN: %f, MAX: %f\n",elem_min,elem_max);
            printf("COORD:%f > %u (elem_min:%f, elem_max:%f)\n",coord,UINT_MAX,elem_min,elem_max);
            terminal_error("Scaling to int failed\n");
        }
#endif
    }
}


FLOAT calc_precision(int D, int z_dimensions, float elem_min, float elem_max, float scale) {
    float precision;
    /*precision = UINT_MAX/(scale); //UINT_MAX=4294967295*/
    precision = UINT_MAX/(1+1*g_rand_scale); //UINT_MAX=4294967295
    float scale_div = D/z_dimensions +1;
    /*printf("scale_div: %f z_dim:%d \n",scale_div,z_dimensions);*/
    precision = precision/(scale_div);

    return precision;
    /*precision = precision*RAND_FLOAT();*/
    /*if(precision < 1){ precision=1;}*/
}


void  generate_z_index
(DataSet* data, int z_dimensions, LONG ** lookup_table, int * zindex, float precision)
{


    vector<float> rand_vector(data->dimensionality);
    vector<int> dimension_permutation(data->dimensionality);

    for(int j = 0; j < data->dimensionality; j++ ) {
        dimension_permutation[j] = j;
    }


    vector<pair<int,uint1024_t>> zindex_tmp(data->size); // TODO: hardcoded uint1024_t

    /*reverse_zindex[i_curve] = new int[data->size];*/
    /*zindex[i_curve] = new int[data->size];*/

        for(int i_dim =0; i_dim < data->dimensionality; i_dim++) {
            rand_vector[i_dim] = RAND_FLOAT()*g_rand_scale;
        }

        random_shuffle(dimension_permutation.begin(), dimension_permutation.end(), rand_int);

#pragma omp parallel for default(shared)
        for (int data_i = 0; data_i < data->size; data_i++) {

            FLOAT* v = get_vector(data,data_i);
            unsigned int P[z_dimensions]; // Each point is projected to this before mapping to z-value

            switch (g_options.projection_type) {
                case SJS:
                    project_SJS(data,v,P,z_dimensions,rand_vector,dimension_permutation,data->elem_min,
                            data->elem_max, precision); break;
            }
                // Generate Z-value for P
            zindex_tmp[data_i] = std::make_pair(data_i,
                    z_value(lookup_table,P,z_dimensions));

            /*zindex_tmp[data_i].second = z_value(lookup_table,P,z_dimensions);*/

         } // END for data_i

        g_timer.tuck("START Z-value based sorting");
#ifdef _OPENMP
        __gnu_parallel::sort(zindex_tmp.begin(), zindex_tmp.end(),dist_compare());
        /*std::sort(zindex_tmp.begin(), zindex_tmp.end(),dist_compare());*/
#else
        std::sort(zindex_tmp.begin(), zindex_tmp.end(),dist_compare());
#endif
        g_timer.tuck("END Z-value based sorting");

        for (int data_i = 0; data_i < data->size ; data_i++) {
            /*reverse_zindex[i_curve][zindex_tmp[data_i].first] = data_i;*/
            zindex[data_i] = zindex_tmp[data_i].first;
        }
}



void debug_zindex(DataSet* data,  int* zindex) {
    for (int data_i = 0; data_i < data->size && data_i < 100; data_i++) {
        printf("%d ",zindex[data_i]);
    }
}


// Parallel version, should be as fast as single thread version on one thread also.
#ifdef _OPENMP

unsigned long long sliding_window_search(DataSet* data, kNNGraph* knng, int* zindex,
        int window_width) {
#define MAXTHREADS 1000
    unsigned long long update_count = 0;
    /*unsigned long long* update_counts = calloc(MAXTHREADS,sizeof(unsigned long long));*/

#pragma omp parallel for default(shared) reduction(+:update_count)
    for(int i_thread = 0; i_thread < g_options.num_threads; i_thread++)
    {
        int numThreads =omp_get_num_threads();
        printf("Num threads:%d\n",numThreads);
        int rangePerThread = floor(data->size/((float)numThreads));
        int rangeStart = rangePerThread*i_thread;
        int rangeEnd = rangePerThread*(i_thread+1);
        if(i_thread==numThreads - 1) {rangeEnd=data->size-1;}
        printf("i_thread:%d rangeStart:%d rangeEnd:%d\n",i_thread,rangeStart,rangeEnd);

        for (int data_i = rangeStart; data_i <= rangeEnd - window_width; data_i++) {
            for (int w = 1; w < window_width ; w++) {
                /*update_counts[i_thread] += update(knng,data,zindex[data_i],zindex[data_i+w]);*/
                update_count += update(knng,data,zindex[data_i],zindex[data_i+w]);
                /*update(knng,data,zindex[data_i],zindex[data_i+w]);*/
             /*update_count += update(knng,data,zindex[data_i],zindex[data_i+w]);*/
             /*g_dist_count++;*/
            }
        }

        /*update_count += update_counts[i_thread];*/
        printf("END THREAD i_thread:%d rangeStart:%d rangeEnd:%d update_count:%llu\n",i_thread,rangeStart,rangeEnd,update_count);
    }
    printf("GLOBAL update_count:%llu\n",update_count);
    /*free(update_counts);*/
    return update_count;
}

#endif

// Single thread version
#ifndef _OPENMP


unsigned long long sliding_window_search(DataSet* data, kNNGraph* knng, int* zindex,
        int window_width) {
    unsigned long long update_count = 0;
#pragma omp parallel for default(shared)
    for (int data_i = 0; data_i < data->size - window_width; data_i++) {
        for (int w = 1; w < window_width ; w++) {
            /*float _dist = distance(DS,zindex[data_i], zindex[data_i+w]);*/
            update_count += update(knng,data,zindex[data_i],zindex[data_i+w]);
            g_dist_count++;
        }
    }
    return update_count;
}


#endif

kNNGraph* z_search(DataSet* data, kNNGraph* knng, int num_curves, int K, int window_width, int z_dimensions) {
    LONG ** lookup_table;
    float precision;
    int * zindex;
    DataSet* projDS = data;
    unsigned long long update_count = 0;
    unsigned long long update_count_nndes = 0;
    float update_portion = 0;
    float update_portion_nndes = 0;
    int *rp;
    int *randpoints;
    vector<vector<UINT> > new_knn(data->size);
    vector<vector<UINT> > old_knn(data->size);

    precision = calc_precision(data->dimensionality, z_dimensions, data->elem_min, data->elem_max,1.1);
    lookup_table = generate_lookup_table(z_dimensions);
    zindex = (int*) safemalloc(sizeof(int)*data->size); //TODO: to UINT

    printf("MIN: %f, MAX: %f, precision:%f\n",data->elem_min,data->elem_max,precision);

    /*if(g_options.randp_distance_project > 0) {*/
    /*projDS = randp_distance_project(data,g_options.randp_distance_project);*/
    /*find_min_max(projDS);*/
    /*}*/
    g_timer.tuck("START generate rand projection points");
    if(g_options.randp_distance_project > 0) {
         rp = generate_heterogenous_rand_points(data, 500, 100);
    }
    g_timer.tuck("END generate rand projection points");
    fflush(stdout);
    randpoints = rp;
    int proj_dim = z_dimensions;
    int nndes_start_k=g_options.recall_K;
    int nndes_max_k=K;
    int nndes_k=nndes_start_k;
    int run_nndes=0;
    int k_increment=2;

    if(g_options.nndes_K > 0) {
        nndes_max_k = nndes_k = g_options.nndes_K;
    }

    find_min_max(projDS);

    for(int i_curve = 0; true ; i_curve++) {
        printf("=====================================================\n");
        /*g_timer.tuck("START iteration");*/
        /*printf("iteration:%d\n",i_curve+1);*/
        printf("START iteration=%d time=%f\n",i_curve+1, g_timer.get_time());

        if(g_options.randp_distance_project > 0) {
            // Experimental
            g_timer.tuck("randp_distance_project");
            proj_dim = g_options.randp_distance_project;
            projDS = randp_distance_project(knng,data,projDS,i_curve,randpoints,proj_dim);
            find_min_max(projDS);
        }

        if(g_options.randp_distance_project > 0) {

            Stat ** stat = (Stat**) safemalloc(sizeof(Stat*)*proj_dim);
            for(int i=0;i<proj_dim;i++) {stat[i]=init_Stat();} //TODO: fix memory leak

            for(int i_dim=0;i_dim<proj_dim;i_dim++) {
                for(int i_data=0;i_data<projDS->size;i_data++) {
                    float val = get_val(projDS, i_data, i_dim);
                    push_Stat(stat[i_dim],val);
                }
            }

            for(int i_dim=0;i_dim<proj_dim;i_dim++) {
                printf("dim:%d ",i_dim);
                debug_Stat(stat[i_dim]);
            }


            for(int i_dim=0;i_dim<proj_dim;i_dim++) {
                //max 32 bit uint: 4294967295
                //max 32 bit int: 2147483647
                if(stat[i_dim]->mean < -0.2 || stat[i_dim]->mean > 0.2)
                {
                    printf("Mean center data for dim:%d \n",i_dim);
                    mean_center_data(projDS,stat[i_dim],i_dim);
                }
            }
            find_min_max(projDS);
        }

        generate_z_index(projDS, proj_dim, lookup_table, zindex, precision);

#ifdef SANITYCHECK
        debug_zindex(projDS,zindex);fflush(stdout);
#endif

        update_count = sliding_window_search(data,knng,zindex,window_width);

        update_portion = ((float)update_count)/((float)(data->size*knng->k)); //TODO:??
        printf("Z-order update count: %u (%f\%)\n",update_count,update_portion);

        /*g_timer.tuck("start calc_recall");*/
        calculate_recall(knng);
        /*g_timer.tuck("end calc_recall");*/

        /*if(i_curve*window_width > K) {*/
        if(g_options.nndes_search && update_portion < g_options.nndes_update_proportion || run_nndes) {
            run_nndes=1;
            g_timer.tuck("start nndes");
            /*update_count_nndes = nndes_iterate(data,knng,new_knn,old_knn,20);*/
            /*update_count_nndes = nndes_iterate(data,knng,new_knn,old_knn,0);*/
            update_count_nndes = nndes_iterate_limited(data,knng,nndes_k);
            update_portion_nndes = ((float)update_count_nndes)/((float)(data->size*K));
            printf("NNDES k=%d update count: %u (%f\%)\n",nndes_k,update_count_nndes,update_portion);
            calculate_recall(knng);
            /*if(update_portion_nndes < 0.02 && nndes_k < nndes_max_k) {nndes_k++;}*/
            if(nndes_k < nndes_max_k) {nndes_k += k_increment; knng->k += k_increment;}
        }

        printf("update_portion:%f update_portion_nndes:%f g_options.delta:%f\n",update_portion, update_portion_nndes, g_options.delta);
        /*if(update_portion < min_update && update_portion_nndes < min_update) { break;}*/
        g_timer.tuck("END iteration");
        if(update_portion + update_portion_nndes < g_options.delta) { break;}


        //TODO:??
        /*if(g_options.randp_distance_project > 0) {*/
        /*free_DataSet(projDS);*/
        /*randpoints++;*/
        /*}*/

        if(g_options.num_iterations > 0 && (i_curve >= g_options.num_iterations - 1)) {
            printf("Iteration limit reached\n",g_options.num_iterations);
            break;
        }
    }
    return knng;
}




//TODO::??
#include "construct_knng_sortv.h"

kNNGraph* construct_knng (DataSet* data, DataSet* DS_proj, int num_curves, int K, int window_width,
        int z_dimensions, int zNN_graph_size, kNNGraph* ground_truth, int use_table) {

    kNNGraph* knng = NULL;
    knng = init_kNNGraph(data->size,g_options.recall_K,K);
    /*find_min_max(data,elem_min,elem_max);*/
    find_min_max(data);

    if(use_table) {
        knng = z_search(data, knng, num_curves, K, window_width, z_dimensions);
    }
    else {
        knng = z_search_sortv(data, knng, num_curves, K, window_width, z_dimensions);
    }

    return knng;
}

void  generate_z_values
(DataSet* data, int z_dimensions, LONG ** lookup_table, int * zindex, float precision, uint1024_t* z_values)
{


    vector<float> rand_vector(data->dimensionality);
    vector<int> dimension_permutation(data->dimensionality);

    for(int j = 0; j < data->dimensionality; j++ ) {
        dimension_permutation[j] = j;
    }


    vector<pair<int,uint1024_t>> zindex_tmp(data->size); // TODO: hardcoded uint1024_t

    /*reverse_zindex[i_curve] = new int[data->size];*/
    /*zindex[i_curve] = new int[data->size];*/

        for(int i_dim =0; i_dim < data->dimensionality; i_dim++) {
            rand_vector[i_dim] = RAND_FLOAT()*g_rand_scale;
        }

        random_shuffle(dimension_permutation.begin(), dimension_permutation.end(), rand_int);

#pragma omp parallel for default(shared)
        for (int data_i = 0; data_i < data->size; data_i++) {

            FLOAT* v = get_vector(data,data_i);
            unsigned int P[z_dimensions]; // Each point is projected to this before mapping to z-value

            switch (g_options.projection_type) {
                case SJS:
                    project_SJS(data,v,P,z_dimensions,rand_vector,dimension_permutation,data->elem_min,
                            data->elem_max, precision); break;
            }
                // Generate Z-value for P
            zindex_tmp[data_i] = std::make_pair(data_i,
                    z_value(lookup_table,P,z_dimensions));
            z_values[data_i] = zindex_tmp[data_i].second;

            /*zindex_tmp[data_i].second = z_value(lookup_table,P,z_dimensions);*/

         } // END for data_i

        g_timer.tuck("START Z-value based sorting");
#ifdef _OPENMP
        __gnu_parallel::sort(zindex_tmp.begin(), zindex_tmp.end(),dist_compare());
        /*std::sort(zindex_tmp.begin(), zindex_tmp.end(),dist_compare());*/
#else
        std::sort(zindex_tmp.begin(), zindex_tmp.end(),dist_compare());
#endif
        g_timer.tuck("END Z-value based sorting");

        for (int data_i = 0; data_i < data->size ; data_i++) {
            /*reverse_zindex[i_curve][zindex_tmp[data_i].first] = data_i;*/
            zindex[data_i] = zindex_tmp[data_i].first;
        }
}

// Calculate proximity preserving quality (Q)
kNNGraph* calc_z_order_stat(DataSet* data, kNNGraph* knng, int num_curves, int K, int window_width, int z_dimensions) {
    LONG ** lookup_table;
    float precision;
    int * zindex;
    DataSet* projDS = data;

    uint1024_t* z_values = (uint1024_t*) safemalloc(sizeof(uint1024_t)*(data->size));

    data->dimensionality = num_curves;


    precision = calc_precision(data->dimensionality, z_dimensions, data->elem_min, data->elem_max,1.1);
    lookup_table = generate_lookup_table(z_dimensions);
    zindex = (int*) safemalloc(sizeof(int)*data->size); //TODO: to UINT
    int* reverse_zindex = (int*) safemalloc(sizeof(int)*data->size); //TODO: to UINT

    printf("MIN: %f, MAX: %f, precision:%f\n",data->elem_min,data->elem_max,precision);

    g_timer.tuck("START generate rand projection points");
    g_timer.tuck("END generate rand projection points");
    fflush(stdout);
    int proj_dim = z_dimensions;

    /*if(g_options.nndes_K > 0) {*/
    /*nndes_max_k = nndes_k = g_options.nndes_K;*/
    /*}*/

    find_min_max(projDS);

    /*double meanQ=0.0;*/
    /*generate_z_index(projDS, proj_dim, lookup_table, zindex, precision);*/
    for(int i_repeat=0;i_repeat<100 ;i_repeat++) {
    generate_z_values(projDS, proj_dim, lookup_table, zindex, precision,z_values);

    for (int data_i = 0; data_i < data->size ; data_i++) {
        reverse_zindex[zindex[data_i]] = data_i;
    }


    int numSample=100000;
    int numErr=0;
    int numOk=0;
    printf("Calculating Z-order stat for DS, Dz=%d\n",z_dimensions);
    for(int i_sample=0;i_sample<numSample;i_sample++) {
        int q = rand_int(data->size);
        int x =q;
        int y =q;
        while(x == q) { x = rand_int(data->size);}
        while(y == q || y == x) { y = rand_int(data->size);}

        double xyorder = 1;

        uint1024_t zqx;
        uint1024_t zqy;
        if(z_values[q] > z_values[x]) {
            zqx = z_values[q] - z_values[x];
        }
        else {
            zqx = z_values[x] - z_values[q];
        }

        if(z_values[q] > z_values[y]) {
            zqy = z_values[q] - z_values[y];
        }
        else {
            zqy = z_values[y] - z_values[q];
        }
        if(zqx > zqy) {xyorder = 1;}
        else {xyorder = -1;}

        double dqx = distance(data, q, x );
        double dqy = distance(data, q, y );
        double corr = (dqx - dqy)*(xyorder);
        int ok = 0;
        if(corr > 0) {ok=1;numOk++;}
        if(corr == 0) {numErr++;}// Equal distance. Should not happen
        /*printf("q:%d x:%d y:%d ok:%d corr:%f\n",q,x,y,ok,corr);*/
    }
    double Q = ((double) numOk)/numSample;
    printf("Q:%f %d/%d numErr:%d\n",Q,numOk,numSample,numErr);
    }
    free(z_values);

    /*exit(0);*/
}


