#ifndef ZVALUE_H_
#define ZVALUE_H_

#include <stdint.h>
#include <limits.h>

#define LOOKUP_SIZE 256
#define LOOKUP_UNIT unsigned int

LONG ** generate_lookup_table(unsigned int dimensions) {

    LONG **lookup_table;

    LONG answer = 0;
    LONG one = 1;

    unsigned int limit = SIZE_LONG/dimensions;
    if(limit > 32) { limit = 32;}

    // Allocate memory
    lookup_table = new LONG*[dimensions];
    for (unsigned int i = 0; i < dimensions; ++i)
        lookup_table[i] = new LONG[LOOKUP_SIZE];

    // Generate lookup table
    for (unsigned int i_dim = 0; i_dim < dimensions; ++i_dim) {
        for (unsigned int j = 0; j < LOOKUP_SIZE; ++j) {
            answer = 0;
            for (unsigned int k = 0; k < limit; ++k) {
                answer |= ((j & (one << k)) << (k*(dimensions-1) + i_dim) );
            }
            lookup_table[i_dim][j] = answer;
        }
    }

    return lookup_table;
}


inline LONG z_value(
        LONG ** lookup_table, unsigned int * point, unsigned int dimensions){

    LONG answer = 0;
    unsigned int rightshift = 32; // == sizeof(unsigned int)*8

    // Once for each 8 bytes in unsigned int
    for(unsigned int i_depth = 1; i_depth <= 4; i_depth++) {
        answer = answer << dimensions*8;
            for(unsigned int i_dim = 0; i_dim < dimensions; i_dim++) {
                answer |= lookup_table[i_dim][
                    (point[i_dim] >> (rightshift -8*i_depth)) & 0xFF ] ;
            }
    }

    return answer;
}

#endif // ZVALUE_H_
